from yellowbox._version import __version__
from yellowbox.clients import docker_client, open_docker_client
from yellowbox.image_build import async_build_image, build_image
from yellowbox.networks import connect, temp_network
from yellowbox.retry import RetrySpec
from yellowbox.service import YellowService
from yellowbox.subclasses import (
    AsyncRunMixin,
    ContainerService,
    RunMixin,
    SingleContainerService,
    SingleEndpointService,
)

__all__ = [
    "__version__",
    "docker_client",
    "open_docker_client",
    "build_image",
    "async_build_image",
    "connect",
    "temp_network",
    "RetrySpec",
    "YellowService",
    "ContainerService",
    "RunMixin",
    "AsyncRunMixin",
    "SingleContainerService",
    "SingleEndpointService",
]



from typing import Any, Dict, Optional

from docker import DockerClient

from yellowbox.containers import create_and_pull_with_defaults
from yellowbox.extras.sql_base import ConnectionOptions, SQLService
from yellowbox.subclasses import SingleContainerService

__all__ = ["MSSQLService"]


class MSSQLService(SQLService, SingleContainerService):
    """
    A postgresSQL service
    """

    INTERNAL_PORT = 1433
    DIALECT = "mssql"

    def __init__(
        self,
        docker_client: DockerClient,
        image="mcr.microsoft.com/mssql/server:latest",
        *,
        admin_password: str = "Swordfish1!",
        product: str = "Developer",
        accept_eula: Optional[str] = None,
        local_driver: Optional[str] = None,
        local_options: Optional[ConnectionOptions] = None,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.admin_password = admin_password
        if accept_eula is None:
            if product == "Developer":
                accept_eula = "y"
            else:
                raise ValueError(f"accept_eula must be set for product {product}")

        if local_driver in (None, "pyodbc") and local_options is None:
            try:
                import pyodbc
            except ImportError as e:
                raise ImportError("pyodbc is required if no other driver is specified") from e
            available_drivers = pyodbc.drivers()
            if not available_drivers:
                raise ValueError(
                    "No odbc drivers found, install drivers here:"
                    " https://docs.microsoft.com/en-us/sql/connect/odbc/"
                    "download-odbc-driver-for-sql-server"
                )
            local_options = {
                "TrustServerCertificate": "yes",
                "driver": available_drivers[0],
            }

        super().__init__(
            create_and_pull_with_defaults(
                docker_client,
                image,
                _kwargs=container_create_kwargs,
                publish_all_ports=True,
                detach=True,
                environment={"ACCEPT_EULA": accept_eula, "SA_PASSWORD": admin_password, "MSSQL_PID": product},
            ),
            default_database="master",
            local_driver=local_driver,
            local_options=local_options,
            **kwargs,
        )

    def userpass(self):
        return "sa", self.admin_password

    def stop(self, signal="SIGINT"):
        # change in default
        return super().stop(signal)



import json
import sys
from asyncio import get_event_loop
from contextlib import asynccontextmanager, contextmanager
from functools import partial
from io import StringIO
from json import JSONDecodeError
from typing import Optional, TextIO
from warnings import warn

from docker import DockerClient
from docker.errors import BuildError, DockerException, ImageNotFound

from yellowbox.utils import _get_spinner


class DockerfileParseError(BuildError):
    pass


DockerfileParseException = DockerfileParseError  # legacy alias


def _docker_build(docker_client: DockerClient, output: TextIO, **kwargs):
    build_log = docker_client.api.build(**kwargs)
    for msg_b in build_log:
        msgs = str(msg_b, "utf-8").splitlines()
        for msg in msgs:
            try:
                parse_msg = json.loads(msg)
            except JSONDecodeError as e:
                raise DockerException("error at build logs") from e
            s = parse_msg.get("stream")
            if s and output:
                print(s, end="", flush=True, file=output)
            else:
                # runtime errors
                error_detail = parse_msg.get("errorDetail")
                # parse errors
                error_msg = parse_msg.get("message")
                # steps of the image creation
                status = parse_msg.get("status")
                # end of process, will contain the ID of the temporary container created at the end
                aux = parse_msg.get("aux")
                if error_detail is not None:
                    raise BuildError(reason=error_detail, build_log=None)
                elif error_msg is not None:
                    raise DockerfileParseError(reason=error_msg, build_log=None)
                elif status is not None and output:
                    print(status, end="", flush=True, file=output)
                elif aux is not None and output:
                    print(aux, end="", flush=True, file=output)
                else:
                    raise DockerException(parse_msg)


@contextmanager
def build_image(
    docker_client: DockerClient,
    image_name: str,
    remove_image: bool = True,
    file: Optional[TextIO] = ...,  # type: ignore[assignment]
    output: Optional[TextIO] = sys.stderr,
    spinner: bool = True,
    **kwargs,
):
    """
    Create a docker image (similar to docker build command)
    At the end, deletes the image (using rmi command)
    Args:
        docker_client: DockerClient to be used to create the image
        image_name: Name of the image to be created. If no tag is provided, the tag "test" will be added.
        remove_image: boolean, whether or not to delete the image at the end, default as True
        file: a file-like object (stream); defaults to the current sys.stderr. if set to None, will disable printing
        spinner: boolean, whether or not to use spinner (default as True), note that this param is set to False in
        case `file` param is not None
    """
    spinner = spinner and file is None
    if file is not ...:
        warn(
            "The `file` parameter is deprecated and will be removed in a future version",
            DeprecationWarning,
            stacklevel=1,
        )
        output = file

    # spinner splits into multiple lines in case stream is being printed at the same time
    if ":" in image_name:
        image_tag = image_name
    else:
        image_tag = f"{image_name}:test"
    yaspin_spinner = _get_spinner(spinner)
    with yaspin_spinner(f"Creating image {image_tag}..."):
        kwargs = {"tag": image_tag, "rm": True, "forcerm": True, **kwargs}
        _docker_build(docker_client, output, **kwargs)
        yield image_tag
        if remove_image:
            try:
                docker_client.api.remove_image(image_tag)
            except ImageNotFound:
                # if the image was already deleted
                pass


@asynccontextmanager
async def async_build_image(
    docker_client: DockerClient,
    image_name: str,
    remove_image: bool = True,
    output: Optional[TextIO] = sys.stderr,
    spinner: bool = True,
    **kwargs,
):
    """
    Create a docker image (similar to docker build command)
    At the end, deletes the image (using rmi command)
    Args:
        docker_client: DockerClient to be used to create the image
        image_name: Name of the image to be created. If no tag is provided, the tag "test" will be added.
        remove_image: boolean, whether or not to delete the image at the end, default as True
        file: a file-like object (stream)
    """
    # spinner splits into multiple lines in case stream is being printed at the same time
    if ":" in image_name:
        image_tag = image_name
    else:
        image_tag = f"{image_name}:test"
    file = StringIO()
    kwargs = {
        "docker_client": docker_client,
        "output": file,
        "tag": image_tag,
        "rm": True,
        "forcerm": True,
        **kwargs,
    }
    if spinner and output:
        print(f".   building image {image_tag}...", file=output)
    try:
        await get_event_loop().run_in_executor(None, partial(_docker_build, **kwargs))
    except (DockerException, BuildError, DockerfileParseError) as e:
        print(f"failed building image {image_tag}", file=sys.stderr)
        print(file.getvalue(), file=sys.stderr)
        raise e
    else:
        if spinner and output:
            print(f".   image {image_tag} built successfully", file=output)
            print(file.getvalue(), file=output)

    yield image_tag
    if remove_image:
        try:
            await get_event_loop().run_in_executor(None, docker_client.api.remove_image, image_tag)
        except ImageNotFound:
            # if the image was already deleted
            pass



from __future__ import annotations

from abc import abstractmethod
from enum import Enum, auto
from typing import TYPE_CHECKING, ContextManager, Mapping, Optional, Tuple, Union

from sqlalchemy import create_engine
from sqlalchemy.exc import InterfaceError, OperationalError
from sqlalchemy_utils import create_database, database_exists, drop_database

from yellowbox import RetrySpec
from yellowbox.containers import get_ports
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleEndpointService
from yellowbox.utils import DOCKER_EXPOSE_HOST, docker_host_name

if TYPE_CHECKING:  # pragma: no cover

    class AsDefault(Enum):
        as_default = auto()

    as_default = AsDefault.as_default
else:
    as_default = object()

ConnectionOptions = Union[str, Mapping[str, str]]


class Database(ContextManager["Database"]):
    # represents a database that is ensured to exist
    def __init__(self, name: str, owner: SQLService):
        self.name = name
        self.owner = owner

    def local_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: Optional[str] = None,
        options: Union[ConnectionOptions, None, AsDefault] = as_default,
    ):
        return self.owner.local_connection_string(dialect, driver, database=self.name, options=options)

    def container_connection_string(
        self,
        hostname: str,
        dialect: Union[str, AsDefault] = as_default,
        driver: Optional[str] = None,
        options: Optional[ConnectionOptions] = None,
    ):
        return self.owner.container_connection_string(hostname, dialect, driver, database=self.name, options=options)

    def host_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: Optional[str] = None,
        options: Optional[ConnectionOptions] = None,
    ):
        return self.owner.host_connection_string(dialect, driver, database=self.name, options=options)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.owner.drop_database(self.name)


def _options_to_string(options: Optional[ConnectionOptions]) -> str:
    if options is None:
        return ""
    if isinstance(options, Mapping):
        return "?" + "&".join(f"{k}={str(v).replace(' ', '+')}" for k, v in options.items())
    if not options.startswith("?"):
        return "?" + options
    return options


class SQLService(
    SingleEndpointService,
    RunMixin,
    AsyncRunMixin,
):
    LOCAL_HOSTNAME = DOCKER_EXPOSE_HOST
    INTERNAL_PORT: int
    DIALECT: str
    DEFAULT_START_RETRYSPEC = RetrySpec(attempts=20)

    def __init__(
        self,
        *args,
        local_driver: Optional[str] = None,
        local_options: Optional[ConnectionOptions] = None,
        default_database: str,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.local_driver = local_driver
        self.local_options = local_options
        self.default_database = default_database

    @abstractmethod
    def userpass(self) -> Tuple[str, str]:
        pass

    def create_database(self, name: str):
        cs = self.local_connection_string(database=name)
        if database_exists(cs):
            raise ValueError(f"Database {name} already exists")
        create_database(cs)

    def drop_database(self, name: str):
        drop_database(self.local_connection_string(database=name))

    def database_exists(self, name: str) -> bool:
        return database_exists(self.local_connection_string(database=name))

    def external_port(self) -> int:
        return get_ports(self._single_endpoint)[self.INTERNAL_PORT]

    def local_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: Union[str, AsDefault, None] = as_default,
        *,
        database: str,
        options: Union[ConnectionOptions, AsDefault, None] = as_default,
    ) -> str:
        """
        Generate an sqlalchemy-style connection string to the database in the service from the docker host.
        Args:
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to.
            options: additional options to pass to the connection string.
        """
        if dialect is as_default:
            dialect = self.DIALECT

        if driver is as_default:
            driver = self.local_driver
        if driver is not None:
            dialect += "+" + driver

        if options is as_default:
            options = self.local_options

        options = _options_to_string(options)

        return (
            f'{dialect}://{":".join(self.userpass())}@{self.LOCAL_HOSTNAME}:{self.external_port()}/'
            f"{database}{options}"
        )

    def container_connection_string(
        self,
        hostname: str,
        dialect: Union[str, AsDefault] = as_default,
        driver: Optional[str] = None,
        *,
        database: str,
        options: Optional[ConnectionOptions] = None,
    ) -> str:
        """
        Generate an sqlalchemy-style connection string to the database in the service from another container on a
         common network.
        Args:
            hostname: the alias of the container.
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to.
            options: additional options to pass to the connection string.
        """
        if dialect is as_default:
            dialect = self.DIALECT

        if driver is not None:
            dialect += "+" + driver

        options = _options_to_string(options)

        return f"{dialect}://{':'.join(self.userpass())}@{hostname}:{self.INTERNAL_PORT}/{database}{options}"

    def host_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: Optional[str] = None,
        *,
        database: str,
        options: Optional[ConnectionOptions] = None,
    ) -> str:
        """
        Generate an sqlalchemy-style connection string to the database in the service from another container.
        Args:
            dialect: The dialect of the sql server.
            driver: additional driver for sqlalchemy to use.
            database: the name of the database to connect to.
            options: additional options to pass to the connection string.
        """
        if dialect is as_default:
            dialect = self.DIALECT

        if driver is not None:
            dialect += "+" + driver

        options = _options_to_string(options)

        return f"{dialect}://{':'.join(self.userpass())}@{docker_host_name}:{self.external_port()}/{database}{options}"

    def database(self, name: str) -> Database:
        if not self.database_exists(name):
            self.create_database(name)
        return Database(name, self)

    def _connect(self):
        cs = self.local_connection_string(database=self.default_database)
        engine = create_engine(cs)
        with engine.connect():
            return

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start(retry_spec)
        retry_spec = retry_spec or self.DEFAULT_START_RETRYSPEC

        retry_spec.retry(self._connect, (OperationalError, InterfaceError))
        return self

    async def astart(self, retry_spec: Optional[RetrySpec] = None) -> None:
        super().start(retry_spec)
        retry_spec = retry_spec or self.DEFAULT_START_RETRYSPEC

        await retry_spec.aretry(self._connect, (OperationalError, InterfaceError))



from contextlib import contextmanager
from typing import Any, Dict, Optional

import aerospike
from aerospike import exception as aerospike_exception
from docker import DockerClient

from yellowbox import RunMixin
from yellowbox.containers import create_and_pull_with_defaults, get_ports
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, SingleContainerService
from yellowbox.utils import DOCKER_EXPOSE_HOST

__all__ = ["AerospikeService", "AEROSPIKE_DEFAULT_PORT"]

AerospikeError = aerospike_exception.AerospikeError  # aerospike doesn't let you import this on its own

AEROSPIKE_DEFAULT_PORT = 3000


class AerospikeService(SingleContainerService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        image="aerospike:ce-6.2.0.3",
        *,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        container = create_and_pull_with_defaults(
            docker_client, image, _kwargs=container_create_kwargs, publish_all_ports=True, detach=True
        )
        self.started = False
        super().__init__(container, **kwargs)

    namespace = "test"

    def client_port(self):
        return get_ports(self.container)[AEROSPIKE_DEFAULT_PORT]

    def _client(self, config: Optional[Dict[str, Any]] = None) -> aerospike.Client:
        config = config or {}
        config = {
            **config,
            "hosts": [(DOCKER_EXPOSE_HOST, self.client_port())],
        }
        return aerospike.client(config)

    @contextmanager
    def client(self, config: Optional[Dict[str, Any]] = None):
        ret = self._client(config)
        try:
            yield ret
        finally:
            ret.close()

    def start(self, retry_spec: Optional[RetrySpec] = None, **kwargs):
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=15)
        # for some reason it takes time before the container is ready to get a connection
        client = retry_spec.retry(self._client, AerospikeError)
        retry_spec.retry(client.is_connected, AerospikeError)
        self.started = True
        return self

    async def astart(self, retry_spec: Optional[RetrySpec] = None, **kwargs) -> None:
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=15)
        client = await retry_spec.aretry(self._client, AerospikeError)
        await retry_spec.aretry(client.is_connected, AerospikeError)
        self.started = True

    def stop(self, signal="SIGKILL"):
        # change in default
        return super().stop(signal)



from __future__ import annotations

import sys
from datetime import datetime
from functools import update_wrapper
from typing import TYPE_CHECKING, AnyStr, Awaitable, Callable, Dict, Iterable, Mapping, Optional, TypeVar, Union

from starlette.requests import Request
from starlette.responses import Response

if TYPE_CHECKING:
    from yellowbox.extras.webserver.endpoints import BASE_HTTP_SIDE_EFFECT, HTTP_SIDE_EFFECT, MockHTTPEndpoint


def reason_is_ne(field: str, expected, got) -> str:
    """
    Create a string that is describes two values being unequal
    Args:
        field: the name of the mismatched field
        expected: the expected value
        got: the actual value
    """
    return f"{field} mismatch: expected {expected}, got {got}"


class MismatchReason(str):
    """
    A falsish object, signifying a failure to match, with a reason.
    """

    def __bool__(self):
        return False


T = TypeVar("T")


def iter_side_effects(side_effects: Iterable[Union[Callable[..., Awaitable[T]], T]]) -> Callable[..., Awaitable[T]]:
    """
    Args:
        side_effects: An iterable of side effects.

    Returns:
        A side effect that varies from call to call. On the first call, delegating to the first side effect on
         side_effects, on the second to the second, and so on.

    Notes:
        This function respects starlette responses in that if it encounters one, it will simply return it instead of
         delegating to it.
    """
    tor = iter(side_effects)

    async def ret(*args, **kwargs):
        next_side_effect = next(tor)
        if isinstance(next_side_effect, Response):  # responses are async callable smh
            return next_side_effect
        return await next_side_effect(*args, **kwargs)

    # we don't want the inner function's __name__ to be the endpoint's name
    ret.__skip_name_for_side_effect__ = True  # type: ignore[attr-defined]

    return ret


def _default_verbose_message_factory(endpoint: MockHTTPEndpoint, request: Request, response: Response) -> str:
    client = f"{request.client.host}:{request.client.port}"
    relative_path = request.url.path
    if request.url.query:
        relative_path += f"?{request.url.query}"
    return (
        f'{datetime.now().isoformat(sep=" ", timespec="seconds")} {endpoint.owner.__name__}:{endpoint.__name__} '
        f"{client} - {request.method} {relative_path} {response.status_code} ({len(response.body)} bytes)"
    )


def verbose_http_side_effect(
    side_effect: BASE_HTTP_SIDE_EFFECT,
    format_function: Callable[[MockHTTPEndpoint, Request, Response], str] = _default_verbose_message_factory,
    file=sys.stdout,
) -> HTTP_SIDE_EFFECT:
    """
    Wrap a side effect so that it prints the arguments and return value on each call.
    """

    def side_effect_factory(endpoint: MockHTTPEndpoint):
        async def side_effect_wrapper(request: Request):
            if isinstance(side_effect, Response):
                response = side_effect
            else:
                response = await side_effect(request)

            message = format_function(endpoint, request, response)
            print(message, file=file)
            return response

        if not isinstance(side_effect, Response):
            update_wrapper(side_effect_wrapper, side_effect)
        else:
            # we don't want the inner function's __name__ to be the endpoint's name
            side_effect_wrapper.__skip_name_for_side_effect__ = True  # type: ignore[attr-defined]
        return side_effect_wrapper

    side_effect_factory.__side_effect_factory__ = True  # type: ignore[attr-defined]
    return side_effect_factory


V = TypeVar("V")


def lower_keys(d: Optional[Mapping[AnyStr, V]]) -> Optional[Dict[AnyStr, V]]:
    return {k.lower(): v for k, v in d.items()} if d else None



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from deprecated import deprecated
from docker import DockerClient
from sqlalchemy import create_engine

from yellowbox.containers import create_and_pull_with_defaults
from yellowbox.extras.sql_base import ConnectionOptions, SQLService, as_default
from yellowbox.subclasses import SingleContainerService

if TYPE_CHECKING:
    from yellowbox.extras.sql_base import AsDefault

__all__ = ["PostgreSQLService", "POSTGRES_INTERNAL_PORT"]

POSTGRES_INTERNAL_PORT = 5432


class PostgreSQLService(SQLService, SingleContainerService):
    """
    A postgresSQL service
    """

    INTERNAL_PORT = POSTGRES_INTERNAL_PORT
    DIALECT = "postgresql"

    def __init__(
        self,
        docker_client: DockerClient,
        image="postgres:latest",
        *,
        user="postgres",
        password="guest",
        default_db: str = None,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Args:
            user: the Name of the default user for the database
            password: The password of the default user for the database
            default_db: The name of the default database. Defaults to the user name.
        """
        if default_db is None:
            default_db = user

        self.user = user
        self.password = password
        self.default_db = default_db
        super().__init__(
            create_and_pull_with_defaults(
                docker_client,
                image,
                _kwargs=container_create_kwargs,
                publish_all_ports=True,
                detach=True,
                environment={"POSTGRES_USER": user, "POSTGRES_PASSWORD": password, "POSTGRES_DB": default_db},
            ),
            default_database=default_db,
            **kwargs,
        )

    def userpass(self):
        return self.user, self.password

    def local_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: Union[str, AsDefault, None] = as_default,
        database: Optional[str] = None,
        options: Union[ConnectionOptions, AsDefault] = as_default,
    ):
        database = database or self.default_db
        return super().local_connection_string(dialect, driver, database=database, options=options)

    def container_connection_string(
        self,
        hostname: str,
        dialect: Union[str, AsDefault] = as_default,
        driver: str = None,
        database: str = None,
        options: ConnectionOptions = None,
    ):
        database = database or self.default_db
        return super().container_connection_string(hostname, dialect, driver, database=database, options=options)

    def host_connection_string(
        self,
        dialect: Union[str, AsDefault] = as_default,
        driver: str = None,
        database: str = None,
        options: ConnectionOptions = None,
    ):
        database = database or self.default_db
        return super().host_connection_string(dialect, driver, database=database, options=options)

    @deprecated(version="0.7.2", reason="Use sqlalchemy.create_engine(service.local_connection_string()) instead")
    def engine(self, **kwargs):
        """
        Create an sqlalchemy Engine connected to the service's default db.
        """
        cs = self.local_connection_string()
        return create_engine(cs, **kwargs)

    @deprecated(
        version="0.7.2", reason="Use sqlalchemy.create_engine(service.local_connection_string()).connect() instead"
    )
    def connection(self, **kwargs):
        """
        Create an sqlalchemy Connection connected to the service's default db.
        """
        return self.engine().connect(**kwargs)

    def stop(self, signal="SIGINT"):
        # change in default
        return super().stop(signal)



from typing import Any, Dict, Optional
from urllib.parse import quote

import requests
from docker import DockerClient
from pika import BlockingConnection, ConnectionParameters, PlainCredentials
from pika.adapters.utils.connection_workflow import AMQPConnectorException
from pika.exceptions import AMQPConnectionError

from yellowbox.containers import create_and_pull_with_defaults, get_ports, upload_file
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleContainerService

__all__ = ["RabbitMQService", "RABBIT_DEFAULT_PORT", "RABBIT_HTTP_API_PORT"]

from yellowbox.utils import DOCKER_EXPOSE_HOST

RABBIT_DEFAULT_PORT = 5672
RABBIT_HTTP_API_PORT = 15672


class RabbitMQService(SingleContainerService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        image="rabbitmq:latest",
        *,
        user="guest",
        password="guest",
        virtual_host="/",
        enable_management=False,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.user = user
        self.password = password
        self.virtual_host = virtual_host
        super().__init__(
            create_and_pull_with_defaults(
                docker_client,
                image,
                _kwargs=container_create_kwargs,
                publish_all_ports=True,
                detach=True,
                ports={RABBIT_HTTP_API_PORT: 0},  # Forward management port by default.
            ),
            **kwargs,
        )

        upload_file(
            self.container,
            "/etc/rabbitmq/rabbitmq.conf",
            f"""
                    default_pass = {password}
                    default_user = {user}
                    default_vhost = {virtual_host}
                    loopback_users = none
                    """.encode(),
        )

        self._enable_management = enable_management

    def connection_port(self):
        return get_ports(self.container)[RABBIT_DEFAULT_PORT]

    def connection(self, **kwargs):
        credentials = PlainCredentials(self.user, self.password)
        connection_params = ConnectionParameters(
            DOCKER_EXPOSE_HOST,
            self.connection_port(),
            credentials=credentials,
            virtual_host=self.virtual_host,
            **kwargs,
        )
        return BlockingConnection(connection_params)

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=30)
        conn = retry_spec.retry(self.connection, (AMQPConnectionError, ConnectionError, AMQPConnectorException))
        conn.close()
        if self._enable_management:
            self.enable_management()
        return self

    async def astart(self, retry_spec: Optional[RetrySpec] = None) -> None:
        super().start()
        retry_spec = retry_spec or RetrySpec(attempts=30)
        conn = await retry_spec.aretry(self.connection, (AMQPConnectionError, ConnectionError, AMQPConnectorException))
        conn.close()
        if self._enable_management:
            self.enable_management()

    def management_url(self):
        try:
            return f"http://{DOCKER_EXPOSE_HOST}:{get_ports(self.container)[RABBIT_HTTP_API_PORT]}/"
        except KeyError as exc:
            raise RuntimeError("Management is not enabled.") from exc

    def enable_management(self):
        if not self.is_alive():
            raise RuntimeError("Must be used on an already-running container.")

        if RABBIT_HTTP_API_PORT not in get_ports(self.container):
            raise RuntimeError("Container must have the management port exposed.")

        self.container.exec_run("rabbitmq-plugins enable rabbitmq_management")

    def reset_state(self, force_queue_deletion=False):
        try:
            management_url = self.management_url()
        except RuntimeError as e:
            raise RuntimeError("management must be enabled for clean_slate") from e

        queues_url = management_url + "api/queues"
        replies = requests.get(queues_url, auth=(self.user, self.password))
        replies.raise_for_status()
        extant_queues = replies.json()
        delete_params = {}
        if not force_queue_deletion:
            delete_params["if-unused"] = "true"
        for queue in extant_queues:
            name = quote(queue["name"], safe="")
            vhost = quote(queue["vhost"], safe="")
            requests.delete(
                management_url + f"api/queues/{vhost}/{name}", auth=(self.user, self.password), params=delete_params
            ).raise_for_status()

    def stop(self, signal="SIGKILL"):
        # change in default
        return super().stop(signal)



import os
import platform
from contextlib import AbstractContextManager, closing, contextmanager, nullcontext
from socket import AF_INET, SO_REUSEADDR, SOCK_STREAM, SOL_SOCKET, socket
from typing import Callable, Optional, TypeVar

from yaspin import yaspin

_T = TypeVar("_T")
_SPINNER_FAILMSG = "💥 "
_SPINNER_SUCCESSMSG = "✅ "


@contextmanager
def _spinner(text):
    with yaspin(text=text) as spinner:
        try:
            yield
        except Exception:
            spinner.fail(_SPINNER_FAILMSG)
            raise
        spinner.ok(_SPINNER_SUCCESSMSG)


def _get_spinner(real=True) -> Callable[[str], AbstractContextManager]:
    if not real:
        return lambda text: nullcontext()
    return _spinner


def get_free_port():
    with closing(socket(AF_INET, SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        return s.getsockname()[1]


uname = platform.uname().release.lower()

if platform.system() == "Linux" and ("microsoft" not in uname):  # catch WSL
    docker_host_name = "172.17.0.1"
else:
    docker_host_name = "host.docker.internal"

# if we expose a port in docker, this is where we expect to find it hosted
# this is almost always just localhost
# you can set the value here to some value with the YELLOWBOX_DOCKER_EXPOSE_HOST env-var
# DO NOT manually change this const, instead call update_docker_expose_host
DOCKER_EXPOSE_HOST = ""


def update_docker_expose_host(value: Optional[str]):
    """
    update the global docker expose host, attempting to infer from the environment

    Args:
        value: if provided, will use this value instead of inferring a host

    Notes:
        This function is called once when yellowbox is imported with a value of the env var
        YELLOWBOX_DOCKER_EXPOSE_HOST

    """
    global DOCKER_EXPOSE_HOST  # noqa: PLW0603

    if value is None:
        value = "127.0.0.1"

    DOCKER_EXPOSE_HOST = value


update_docker_expose_host(os.getenv("YELLOWBOX_DOCKER_EXPOSE_HOST"))



from __future__ import annotations

import builtins
from dataclasses import dataclass
from enum import Enum, auto
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
    overload,
)

from starlette.requests import HTTPConnection
from starlette.websockets import WebSocket

from yellowbox.extras.webserver.request_capture import ScopeExpectation
from yellowbox.extras.webserver.util import MismatchReason, reason_is_ne


class WebSocketRecorder(WebSocket):
    """
    A subclass of starlette.Websocket that records the conversation in a transcript
    messages are parsed as per https://asgi.readthedocs.io/en/latest/specs/www.html#websocket
    """

    def __init__(self, scope, receive, send, sinks: Sequence[RecordedWSTranscripts]):
        """
        Args:
            scope: forwarded to Websocket
            receive: forwarded to Websocket
            send: forwarded to Websocket
            sinks: a list of transcripts to add the transcript to
        """
        super().__init__(scope, receive, send)
        self.transcript: RecordedWSTranscript = RecordedWSTranscript.from_connection(self)
        for sink in sinks:
            sink.append(self.transcript)

    async def receive(self):
        message = await super().receive()
        message_type = message.get("type")
        if message_type == "websocket.receive":
            # by asgi uvicorn implementation (def asgi_receive), a message will always contain exactly one of
            #  'text' or 'bytes'
            data = message.get("text")
            if data is None:
                data = message.get("bytes")
            self.transcript.append(RecordedWSMessage(data, Sender.Client))
        elif message_type == "websocket.disconnect":
            self.transcript.close = (Sender.Client, message.get("code", 1000))
        return message

    async def send(self, message):
        message_type = message["type"]
        if message_type == "websocket.accept":
            self.transcript.accepted = True
        elif message_type == "websocket.close":
            self.transcript.close = (Sender.Server, message.get("code", 1000))
        elif message_type == "websocket.send":
            # by asgi uvicorn implementation (def asgi_receive), a message will always contain exactly one of
            #  'text' or 'data'
            data = message.get("text")
            if data is None:
                data = message.get("bytes")
            self.transcript.append(RecordedWSMessage(data, Sender.Server))
        await super().send(message)


class RecorderEndpoint:
    """
    An entrypoint app that records websocket conversations
    Args:
        function: the function to call the endpoint with
        sinks: a list of transcripts that the record should be added to
    """

    def __init__(self, function, sinks: Sequence[RecordedWSTranscripts]):
        self.function = function
        self.sinks = sinks

    async def __call__(self, scope, receive, send):
        ws = WebSocketRecorder(scope, receive, send, sinks=self.sinks)
        return await self.function(ws)


class Sender(Enum):
    """
    A sender that can send messages in a transcript. A sender can be called to create an expected message.

    Examples:
        >>> import re
        ... expected_message = Sender.Server(re.compile('[a-z][a-z][0-9]'))
        ... message = RecordedWSMessage('tk1', Sender.Server)
        ... assert expected_message.matches(message)
    """

    Server = auto()
    Client = auto()

    def __call__(self, data: Union[Pattern[str], Pattern[bytes], str, bytes, builtins.ellipsis]) -> ExpectedWSMessage:
        """
        Create an expected message, originating from this caller
        Args:
            data: the expected data of the message. Can be either a bytes or string for an exact match, a compiled
             pattern for a full match, or ellipsis to match any data.
        """
        return ExpectedWSMessage(data, self)


@dataclass
class RecordedWSMessage:
    """
    A recorded websocket message, with a sender and value sent
    """

    data: Union[bytes, str]
    sender: Sender

    def __repr__(self):
        return f"RecordedWSMessage({self.data!r}, Sender.{self.sender.name})"


class RecordedWSTranscript(List[RecordedWSMessage]):
    """
    A transcript of a single websocket connection
    """

    def __init__(
        self,
        arg: Iterable[RecordedWSMessage],
        headers: Dict[str, List[str]],
        path: str,
        path_params: Dict[str, Any],
        query_params: Dict[str, List[str]],
    ):
        """
        Args:
            arg: forwarded to super (List)
            headers: the headers of the connection
            path: the path of the connection request
            path_params: the path params of the connection request, as specified by the route's starlette rule string
            query_params: the query params of the connection request
        """
        super().__init__(arg)
        self.headers = headers
        self.path = path
        self.path_params = path_params
        self.query_params = query_params
        self.accepted: bool = False
        """Whether the message was ever accepted by the server"""
        self.close: Optional[Tuple[Sender, int]] = None
        """The sender who closed the connection, along with the close code"""

    @classmethod
    def from_connection(cls, connection: HTTPConnection):
        """
        Create an empty transcript from a new http connection
        Args:
            connection: the http connection to use
        """
        headers: Dict[str, List[str]] = {}
        for h_k, h_v in connection.headers.items():
            h_k_lower = h_k.lower()
            h_lst = headers.get(h_k_lower)
            if h_lst is None:
                headers[h_k_lower] = [h_v]
            else:
                h_lst.append(h_v)
        path = connection.url.path
        path_params = connection.path_params
        query_params: Dict[str, List[str]] = {}
        for q_k, q_v in connection.query_params.multi_items():
            q_lst = query_params.get(q_k)
            if q_lst is None:
                query_params[q_k] = [q_v]
            else:
                q_lst.append(q_v)
        return cls((), headers, path, path_params, query_params)


class ExpectedWSMessage:
    """
    A single expected message in a websockets connection
    """

    def __init__(self, data: Union[Pattern[str], Pattern[bytes], str, bytes, builtins.ellipsis], sender: Sender):
        """
        Args:
            data: the expected data. Can be either a bytes or string for an exact match, a compiled pattern for a full
             match, or ellipsis to match any data.
            sender: The sender to expect the message from.
        """
        self.data = data
        self.sender = sender

    def matches(self, message: RecordedWSMessage) -> Union[bool, MismatchReason]:
        """
        Checks whether a recorded message matches the expectation
        Args:
            message: the recorded message

        Returns:
            True if the message matches, or a mismatch reason otherwise.
        """
        if self.sender != message.sender:
            return MismatchReason(reason_is_ne("sender", self.sender, message.sender))

        if self.data is ...:
            return True
        elif isinstance(self.data, Pattern):
            try:
                match = self.data.fullmatch(message.data)  # type:ignore[arg-type]
            except TypeError:  # this would happen when we try to use byte patterns on strs or vice-versa
                match = None
            if not match:
                return MismatchReason(f"expected data to match pattern {self.data.pattern!r}, got {message.data!r}")
            return True
        elif self.data != message.data:
            return MismatchReason(reason_is_ne("data", self.data, message.data))
        return True

    def __str__(self):
        return f"{self.sender.name}({self.data})"


class ExpectedWSTranscript(ScopeExpectation):
    """
    An expected websocket transcript
    """

    def __init__(
        self,
        messages: Sequence[Union[builtins.ellipsis, ExpectedWSMessage]] = (...,),
        headers: Optional[Mapping[str, Collection[str]]] = None,
        headers_submap: Optional[Mapping[str, Collection[str]]] = None,
        path: Optional[Union[str, Pattern[str]]] = None,
        path_params: Optional[Mapping[str, Any]] = None,
        path_params_submap: Optional[Mapping[str, Any]] = None,
        query_params: Optional[Mapping[str, Collection[str]]] = None,
        query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
        close: Optional[Tuple[Sender, int]] = None,
        accepted: Optional[bool] = True,
    ):
        """
        Args:
            messages: the expected messages in the transcript, can begin or end ellipsis to signify that any
                number of messages can precede or follow the messages to match.
            headers: If specified, expects the request to have these headers exactly
            headers_submap: If specified expects the request to have at least the headers specified
            path: If specified, expected the request url (after the host) to either be equal to this (in case of string)
             or to fully match the regex pattern provided
            path_params: If specified, expects the request to have exactly these path parameters, as provided by the
             starlette request
            path_params_submap: if specified, expects the result to have at least theses path parameters, as provided
             by the starlette request
            query_params: If specified, expects the request to have these query arguments exactly
            query_params_submap: If specified, expects the request to have at least these query arguments
            close: If specified, must be a tuple of sender and an integer. Expects the transcript to be closed by that
             sender with that code.
            accepted: If specified, will only match transcripts of accepted connections (if True) or rejected
             connections (if False)
        """
        super().__init__(
            headers, headers_submap, path, path_params, path_params_submap, query_params, query_params_submap
        )
        if messages and messages[0] is ...:
            messages = messages[1:]
            self.any_start = True
        else:
            self.any_start = False

        if messages and messages[-1] is ...:
            messages = messages[:-1]
            self.any_end = True
        else:
            self.any_end = False

        if any(m is ... for m in messages):
            raise TypeError("ExpectedWSTranscript can only contain ellipsis at the start and at the end")

        self.expected_messages: Tuple[ExpectedWSMessage] = messages  # type:ignore[assignment]
        self.accepted = accepted
        self.close = close

    def matches(self, recorded: RecordedWSTranscript):
        """
        Checks whether a recorded transcript matches the expectation
        Args:
            recorded: the recorded transcript

        Returns:
            True if the transcript matches, or a mismatch reason otherwise.
        """
        if recorded.close is None:
            raise RuntimeError("the transcript is not yet done")

        reasons = list(self.scope_mismatch_reasons(recorded))

        if self.close is not None and recorded.close != self.close:
            reasons.append(reason_is_ne("close_code", self.close, recorded.close))
        if self.accepted is not None and recorded.accepted != self.accepted:
            reasons.append(reason_is_ne("accepted", self.accepted, recorded.accepted))

        if not self.any_start and not self.any_end and len(recorded) != len(self.expected_messages):
            reasons.append(f"expected exactly {len(self.expected_messages)} messages, found {len(recorded)}")
        if len(recorded) < len(self.expected_messages):
            reasons.append(f"expected at least {len(self.expected_messages)} messages, found only {len(recorded)}")
        if reasons:
            return MismatchReason(", ".join(reasons))
        # in order to account for any_start, any_end, we check every subsequence of the recorded transcript
        # we store a list of candidate indices for the subsequence start
        if not self.any_start:
            # the case of (not any_start and not any_end) is covered by the first condition
            indices: Iterable[int] = (0,)
        elif not self.any_end:
            indices = (len(recorded) - len(self.expected_messages),)
        else:
            indices = range(0, len(recorded) - len(self.expected_messages) + 1)

        whynots = []
        for start_index in indices:
            for i, expected_message in enumerate(self.expected_messages):
                rec = recorded[start_index + i]
                match = expected_message.matches(rec)
                if not match:
                    whynots.append(f"{rec} did not match {expected_message}: {match}")
                    break
            else:
                return True

        return MismatchReason("could not find expected calls" + "".join("\n\t" + wn for wn in whynots))


class RecordedWSTranscripts(List[RecordedWSTranscript]):
    """
    A list of recorded WS transcripts, in the order that the connections began
    """

    def assert_not_requested(self):
        """
        asserts that no connections were recorded.
        """
        if self:
            raise AssertionError(f"{len(self)} requests were made")

    def assert_requested(self):
        """
        asserts that at least one transcript was recorded.
        """
        if not self:
            raise AssertionError("No requests were made")

    def assert_requested_once(self):
        """
        asserts that exactly one transcript was recorded.
        """
        if not self:
            raise AssertionError("No requests were made")
        if len(self) > 1:
            raise AssertionError(f"{len(self)} requests were made")

    @overload
    def assert_requested_with(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_requested_with(
        self,
        *,
        messages: Sequence[Union[builtins.ellipsis, ExpectedWSMessage]] = (...,),
        headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
        headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
        path: Optional[Union[str, Pattern[str]]] = None,
        path_params: Optional[Mapping[str, Any]] = None,
        path_params_submap: Optional[Mapping[str, Any]] = None,
        query_params: Optional[Mapping[str, Collection[str]]] = None,
        query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
        close: Optional[Tuple[Sender, int]] = None,
        accepted: Optional[bool] = True,
    ):
        ...

    def assert_requested_with(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that the latest transcript recorded matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError("method can be called with either expected or keyword args, but not both")
        if not expected:
            if not kwargs:
                raise TypeError("either expected or keyword args must be provided")
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError("No requests were made")
        match = expected.matches(self[-1])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_requested_once_with(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_requested_once_with(
        self,
        *,
        messages: Sequence[Union[builtins.ellipsis, ExpectedWSMessage]] = (...,),
        headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
        headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
        path: Optional[Union[str, Pattern[str]]] = None,
        path_params: Optional[Mapping[str, Any]] = None,
        path_params_submap: Optional[Mapping[str, Any]] = None,
        query_params: Optional[Mapping[str, Collection[str]]] = None,
        query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
        close: Optional[Tuple[Sender, int]] = None,
        accepted: Optional[bool] = True,
    ):
        ...

    def assert_requested_once_with(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that there is only one transcript, and that it matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError("method can be called with either expected or keyword args, but not both")
        if not expected:
            if not kwargs:
                raise TypeError("either expected or keyword args must be provided")
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError("No requests were made")
        if len(self) > 1:
            raise AssertionError(f"{len(self)} requests were made")
        match = expected.matches(self[0])
        if not match:
            raise AssertionError(str(match))

    @overload
    def assert_any_request(self, expected: ExpectedWSTranscript):
        ...

    @overload
    def assert_any_request(
        self,
        *,
        messages: Sequence[Union[builtins.ellipsis, ExpectedWSMessage]] = (...,),
        headers: Optional[Mapping[bytes, Collection[bytes]]] = None,
        headers_submap: Optional[Mapping[bytes, Collection[bytes]]] = None,
        path: Optional[Union[str, Pattern[str]]] = None,
        path_params: Optional[Mapping[str, Any]] = None,
        path_params_submap: Optional[Mapping[str, Any]] = None,
        query_params: Optional[Mapping[str, Collection[str]]] = None,
        query_params_submap: Optional[Mapping[str, Collection[str]]] = None,
        close: Optional[Tuple[Sender, int]] = None,
        accepted: Optional[bool] = True,
    ):
        ...

    def assert_any_request(self, expected: Optional[ExpectedWSTranscript] = None, **kwargs):
        """
        Asserts that the at least one transcript recorded matches an expected transcript
        Args:
            expected: an expected transcript.
            **kwargs: if an expected request is not provided, then a new expected request is constructed by forwarding
             the keyword arguments to the constructor of ExpectedWSTranscript.
        """
        if expected and kwargs:
            raise TypeError("method can be called with either expected or keyword args, but not both")
        if not expected:
            if not kwargs:
                raise TypeError("either expected or keyword args must be provided")
            expected = ExpectedWSTranscript(**kwargs)

        if not self:
            raise AssertionError("No requests were made")
        if any(expected.matches(req) for req in self):
            return
        raise AssertionError("No transcripts match")



from contextlib import closing, contextmanager

from docker import DockerClient


@contextmanager
def open_docker_client():
    """
    Starts docker client from the environment, with a fallback to default TCP port
    (for running from within virtual machines)
    """
    try:
        ret = DockerClient.from_env()
        ret.ping()
    except Exception:  # noqa: BLE001
        ret = DockerClient(base_url="tcp://localhost:2375")
        ret.ping()
    with closing(ret):
        yield ret


# legacy alias
docker_client = open_docker_client



from typing import IO, Any, Callable, ContextManager, Dict, Mapping, Optional, Sequence, TypeVar, Union, overload

from docker import DockerClient
from redis import ConnectionError as RedisConnectionError, Redis

from yellowbox import RunMixin
from yellowbox.containers import create_and_pull_with_defaults, get_ports, upload_file
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, SingleContainerService

__all__ = ["RedisService", "REDIS_DEFAULT_PORT", "DEFAULT_RDB_PATH", "append_state"]

from yellowbox.utils import DOCKER_EXPOSE_HOST

REDIS_DEFAULT_PORT = 6379
DEFAULT_RDB_PATH = "/data/dump.rdb"

_T = TypeVar("_T", bound=ContextManager)
RedisPrimitive = Union[str, int, float, bytes]
RedisState = Mapping[str, Union[RedisPrimitive, Mapping[Union[str, bytes], RedisPrimitive], Sequence[RedisPrimitive]]]


def append_state(client: Redis, db_state: RedisState):
    for k, v in db_state.items():
        if isinstance(v, Sequence):
            client.rpush(k, *v)
        elif isinstance(v, Mapping):
            client.hset(k, mapping=v)
        else:
            client.set(k, v)


class RedisService(SingleContainerService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        image="redis:latest",
        redis_file: Optional[IO[bytes]] = None,
        *,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        container = create_and_pull_with_defaults(
            docker_client, image, _kwargs=container_create_kwargs, publish_all_ports=True, detach=True
        )
        self.started = False
        super().__init__(container, **kwargs)

        if redis_file:
            self.set_rdb(redis_file)

    def set_rdb(self, redis_file: IO[bytes]):
        if self.started:
            raise RuntimeError("Server already started. Cannot set RDB.")
        upload_file(self.container, DEFAULT_RDB_PATH, fileobj=redis_file)

    def client_port(self):
        return get_ports(self.container)[REDIS_DEFAULT_PORT]

    @overload
    def client(self, *, client_cls: Callable[..., _T], **kwargs) -> _T:
        ...

    @overload
    def client(self, **kwargs) -> Redis:
        ...

    def client(self, *, client_cls=Redis, **kwargs) -> Any:
        port = self.client_port()
        return client_cls(host=DOCKER_EXPOSE_HOST, port=port, **kwargs)

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        client_cm: ContextManager[Redis] = self.client()
        with client_cm as client:
            retry_spec = retry_spec or RetrySpec(attempts=15)
            retry_spec.retry(client.ping, RedisConnectionError)
        self.started = True
        return self

    async def astart(self, retry_spec: Optional[RetrySpec] = None) -> None:
        super().start()
        client_cm: ContextManager[Redis] = self.client()
        with client_cm as client:
            retry_spec = retry_spec or RetrySpec(attempts=15)
            await retry_spec.aretry(client.ping, RedisConnectionError)
        self.started = True

    def reset_state(self) -> None:
        client: Redis
        with self.client() as client:
            client.flushall()

    def set_state(self, db_dict: RedisState) -> None:
        client_cm: ContextManager[Redis] = self.client()
        with client_cm as client:
            client.flushall()
            append_state(client, db_dict)



from __future__ import annotations

import io
import os
import shutil
import stat
import tarfile
from contextlib import contextmanager
from os import PathLike
from tempfile import TemporaryFile
from typing import (
    IO,
    Any,
    Container as AbstractContainer,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

import docker
from docker import DockerClient
from docker.errors import ImageNotFound
from docker.models.containers import Container
from docker.models.networks import Network
from requests import HTTPError

__all__ = [
    "get_ports",
    "get_aliases",
    "is_alive",
    "is_removed",
    "killing",
    "create_and_pull",
    "download_file",
    "upload_file",
    "SafeContainerCreator",
    "removing",
]

_DEFAULT_TIMEOUT = 10

_T = TypeVar("_T")
_CT = TypeVar("_CT", bound=Container)


def get_ports(container: Container) -> Dict[int, int]:
    """Get the exposed (published) ports of a given container

    Useful for when the ports are assigned dynamically.

    Example:
        >>> c = Container("redis:latest", publish_all_ports=True)
        >>> c.run()
        >>> c.reload()  # Must reload to get updated config.
        >>> ports = get_ports(c)
        >>> ports[6379]  # Random port assigned to 6379 inside the container
        1234

    Note: Container must be up and running. To make sure data is up-to-date,
    make sure you .reload() the container before attempting to fetch the ports.

    Args:
        container: Docker container.

    Returns:
        Port mapping {internal_container_port: external_host_port}.
    """
    # todo this probably won't work for multi-network containers
    ports = {}
    portmap = container.attrs["NetworkSettings"]["Ports"]
    for port, external_address in portmap.items():
        # Filter out unpublished ports.
        if external_address is None:
            continue

        assert len(external_address) > 0

        external_port = int(external_address[0]["HostPort"])

        port_num, *_ = port.partition("/")  # Strip out type (tcp, udp, ...)
        ports[int(port_num)] = int(external_port)

    return ports


def get_aliases(container: Container, network: Union[str, Network]) -> Sequence[str]:
    if not isinstance(network, str):
        network = network.name
    return container.attrs["NetworkSettings"]["Networks"][network]["Aliases"]


def short_id(container: Container) -> str:
    """Get the short 12-character id of a container

    By default, the short 12-character id is used as a network alias in all
    of the networks connected to the container.

    Args:
        container: Docker container to retrieve ID from.

    Returns:
        12 character string.
    """
    return container.id[:12]


def is_alive(container: Container) -> bool:
    if is_removed(container):
        return False
    return container.status.lower() not in ("exited", "stopped")


@contextmanager
def killing(
    container: _CT, *, timeout: float = _DEFAULT_TIMEOUT, signal: str = "SIGKILL"
) -> Generator[_CT, None, None]:
    """A context manager that kills a docker container upon completion.

    Example:
        container = DockerContainer(...)
        with killing(container):
            ...
        # Container is now killed with SIGKILL

    Args:
        container: Container to kill upon completion
        timeout: Time to wait for container to be killed after sending a signal.
        Defaults to 10 seconds.
        signal: The signal to send to the container.

    Returns:
        A context manager to be used in a 'with' statement.
    """
    try:
        yield container
    finally:
        if is_alive(container):
            container.kill(signal)
            container.wait(timeout=timeout)


@contextmanager
def removing(
    container: _CT, *, expected_exit_code: Optional[Union[int, AbstractContainer[int]]] = 0, force=False
) -> Generator[_CT, None, None]:
    """A context manager that removes a docker container upon completion.

    Example:
        container = DockerContainer(...)
        with removing(container):
            ...
        # Container is now removed

    Args:
        container: Container to remove upon completion
        expected_exit_code: Expected exit code or codes of the container. If the container exited with a different
            code, an exception will be raised. Defaults to 0.
        force: If True, the container will be stopped and removed even if it is running.

    Returns:
        A context manager to be used in a 'with' statement.
    """
    try:
        yield container
    finally:
        if not is_removed(container):
            # checking if the container is removed reloads the container
            if is_alive(container):
                if not force:
                    raise RuntimeError(f"Container {container.id} is still alive (status: {container.status})")
                container.kill("SIGKILL")
            result = container.wait(timeout=10)
            if expected_exit_code is not None:
                if isinstance(expected_exit_code, int):
                    expected_exit_code = (expected_exit_code,)

                if result["StatusCode"] not in expected_exit_code:
                    raise RuntimeError(f"Container {container.id} exited with code {result['StatusCode']}")
            container.remove(force=force, v=True)


def create_and_pull(docker_client: DockerClient, image: str, *args, _kwargs=None, **kwargs) -> Container:
    """
    Create a docker container, pulling the image if necessary.
    Args:
        docker_client: the docker client to use.
        image: the image name to create
        *args: additional arguments forwarded to ``docker_client.containers.create``
        **kwargs: additional arguments forwarded to ``docker_client.containers.create``

    Returns:
        A non-started container.

    Note:
        Due to inconsistent behaviour of docker's "pull" command across
        platforms, this function will raise an error if no tag is specified
    """
    name, _, tag = image.partition(":")
    if not tag:
        raise ValueError("the image name must contain a tag")
    try:
        ret = docker_client.containers.create(image, *args, **kwargs)
    except ImageNotFound:
        docker_client.images.pull(image, platform=None)
        ret = docker_client.containers.create(image, *args, **kwargs)
    return ret


def create_and_pull_with_defaults(*args, _kwargs: Optional[Dict[str, Any]] = None, **default_kwargs):
    if _kwargs:
        kwargs = {**default_kwargs, **_kwargs}
    else:
        kwargs = default_kwargs

    return create_and_pull(*args, **kwargs)


def is_removed(container: Container):
    try:
        container.reload()
    except HTTPError:
        return True
    return False


def download_file(container: Container, path: Union[str, PathLike]) -> IO[bytes]:
    """Download a file from the given container

    Args:
        container: Docker container at any state.
        path: File path.

    Raises:
        FileNotFoundError: Path was not found.
        IsADirectoryError: Path is not a regular file.
    """
    realpath = os.fspath(path)
    exc: Exception
    try:
        iterator, stats = container.get_archive(realpath, chunk_size=None)
    except docker.errors.NotFound:
        exc = FileNotFoundError(realpath)
        exc.filename = realpath
        raise exc

    if stat.S_ISDIR(stats["mode"]):
        exc = IsADirectoryError(path)
        exc.filename = realpath
        raise exc

    # Finalizer ensures temporary file will close and be removed.
    temp_file = TemporaryFile("w+b")

    for chunk in iterator:
        temp_file.write(chunk)
    temp_file.seek(0)

    tar_file = tarfile.open(fileobj=temp_file)
    member = tar_file.next()
    return cast("IO[bytes]", tar_file.extractfile(member))


def upload_file(
    container: Container, path: Union[str, PathLike[str]], data: bytes = None, fileobj: IO[bytes] = None
) -> None:
    """Upload a file to the given container

    Args:
        container: Docker container.
        path: Path to upload the file to.
        data: Bytes of data to upload. Cannot be set with fileobj.
        fileobj: File object to uplaod. Cannot be set with data.
    """
    if data is fileobj is None:
        raise TypeError("data or fileobj must be set.")

    if data is not None is not fileobj:
        raise TypeError("Can't set both data and fileobj.")

    filename = os.path.basename(path)

    tar_data = _create_tar(filename, data, fileobj)

    container.put_archive(os.path.dirname(path), tar_data)


def _create_tar(filename, data=None, fileobj=None) -> bytes:
    """Create a tarfile made of the given data

    Args:
        filename: Name of the file to create inside the tar
        data: Data of the file. Cannot exist with fileobj.
        fileobj: File object. Cannot exist with data.

    Returns:
        Bytes of a tarfile, containing the given file.
    """
    output = io.BytesIO()
    with tarfile.open(fileobj=output, mode="w") as tar:
        if data is not None:
            tarinfo = tarfile.TarInfo(name=filename)
            tarinfo.size = len(data)
            tar.addfile(tarinfo, io.BytesIO(data))
        else:
            try:
                # Attempt to extract info (such as size) from file object
                tarinfo = tar.gettarinfo(arcname=filename, fileobj=fileobj)
                tar.addfile(tarinfo, fileobj)
            except (OSError, AttributeError):
                # Failed to extract info, writing and reading from temp file.
                with TemporaryFile("w+b") as temp_file:
                    shutil.copyfileobj(fileobj, temp_file)
                    temp_file.seek(0)
                    temp_file.flush()
                    tarinfo = tar.gettarinfo(arcname=filename, fileobj=temp_file)
                    tar.addfile(tarinfo, temp_file)
    return output.getvalue()


class SafeContainerCreator:
    """
    A class that can safely pull and create multiple containers in succession, where if one fails, all the previous
     ones are removed
    """

    def __init__(self, client: DockerClient):
        self.client = client
        self.created: List[Container] = []

    def create_and_pull(self, image, command=None, **kwargs):
        try:
            container = create_and_pull(self.client, image, command, **kwargs)
        except Exception:
            for container in reversed(self.created):
                container.remove(v=True)
            raise
        self.created.append(container)
        return container



from docker import DockerClient
from pytest import fixture

from yellowbox.clients import open_docker_client


@fixture(scope="session")
def docker_client() -> DockerClient:
    with open_docker_client() as client:
        yield client



from yellowbox.extras.webserver.class_endpoint import class_http_endpoint, class_ws_endpoint
from yellowbox.extras.webserver.endpoints import MockHTTPEndpoint, MockWSEndpoint, http_endpoint, ws_endpoint
from yellowbox.extras.webserver.http_request_capture import ExpectedHTTPRequest
from yellowbox.extras.webserver.util import iter_side_effects, verbose_http_side_effect
from yellowbox.extras.webserver.webserver import WebServer
from yellowbox.extras.webserver.ws_request_capture import ExpectedWSTranscript, Sender

__all__ = [
    "WebServer",
    "http_endpoint",
    "ws_endpoint",
    "ExpectedHTTPRequest",
    "ExpectedWSTranscript",
    "Sender",
    "MockHTTPEndpoint",
    "MockWSEndpoint",
    "class_ws_endpoint",
    "class_http_endpoint",
    "iter_side_effects",
    "verbose_http_side_effect",
]



from contextlib import contextmanager
from os import environ, getenv
from typing import Any, Dict, Iterable, Optional
from warnings import warn

import requests
from docker import DockerClient

from yellowbox.containers import create_and_pull_with_defaults, get_ports
from yellowbox.retry import RetrySpec
from yellowbox.subclasses import AsyncRunMixin, RunMixin, SingleContainerService
from yellowbox.utils import DOCKER_EXPOSE_HOST, docker_host_name

FAKE_GCS_DEFAULT_PORT = 4443


class FakeGoogleCloudStorage(SingleContainerService, RunMixin, AsyncRunMixin):
    def __init__(
        self,
        docker_client: DockerClient,
        image: str = "fsouza/fake-gcs-server:1.42",
        scheme: str = "https",
        command: str = "",
        *,
        container_create_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        # note that fake-gcs-server 1.43.0 has a bug https://github.com/fsouza/fake-gcs-server/issues/1034
        command = f"-scheme {scheme} {command}"
        self.scheme = scheme
        container = create_and_pull_with_defaults(
            docker_client, image, command, _kwargs=container_create_kwargs, publish_all_ports=True
        )
        super().__init__(container, **kwargs)

    def client_port(self):
        return get_ports(self.container)[FAKE_GCS_DEFAULT_PORT]

    def local_url(self, scheme: Optional[str] = ...):  # type: ignore[assignment]
        ret = f"{DOCKER_EXPOSE_HOST}:{self.client_port()}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def container_url(self, hostname: str, scheme: Optional[str] = ...):  # type: ignore[assignment]
        ret = f"{hostname}:{FAKE_GCS_DEFAULT_PORT}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def host_url(self, scheme: Optional[str] = ...):  # type: ignore[assignment]
        ret = f"{docker_host_name}:{self.client_port()}"
        if scheme is ...:
            scheme = self.scheme
        if scheme:
            ret = f"{self.scheme}://{ret}"
        return ret

    def start(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        url = self.local_url() + "/storage/v1/b"
        retry_spec = retry_spec or RetrySpec(attempts=15)
        retry_spec.retry(
            lambda: requests.get(url, verify=False).raise_for_status(), requests.exceptions.RequestException
        )
        return self

    async def astart(self, retry_spec: Optional[RetrySpec] = None):
        super().start()
        url = self.local_url() + "/storage/v1/b"
        retry_spec = retry_spec or RetrySpec(attempts=15)
        await retry_spec.aretry(
            lambda: requests.get(url, verify=False).raise_for_status(), requests.exceptions.RequestException
        )

    @contextmanager
    def patch_gcloud_aio(self):
        from gcloud.aio.storage import __version__ as gcloud_aio_version

        if not gcloud_aio_version.startswith("7."):
            # for newer gcloud_aio, we can just adjust the environment
            warn(
                "newer gcloud versions should be patched directly by setting the environment variable "
                "STORAGE_EMULATOR_HOST to service.local_url()",
                stacklevel=1,
            )
            prev_env = getenv("STORAGE_EMULATOR_HOST")
            environ["STORAGE_EMULATOR_HOST"] = self.local_url(None)
            yield
            if prev_env is None:
                del environ["STORAGE_EMULATOR_HOST"]
            else:
                environ["STORAGE_EMULATOR_HOST"] = prev_env
            return
        import gcloud.aio.storage.storage as gcloud_module

        previous_state = (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        )
        (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        ) = (
            self.local_url() + "/storage/v1/b",
            self.local_url() + "/upload/storage/v1/b",
            False,
            self.local_url(scheme=""),
        )
        yield
        (
            gcloud_module.API_ROOT,
            gcloud_module.API_ROOT_UPLOAD,
            gcloud_module.VERIFY_SSL,
            gcloud_module.STORAGE_EMULATOR_HOST,
        ) = previous_state

    def create_bucket(self, bucket_name: str) -> Dict[str, Any]:
        url = self.local_url()

        resp = requests.post(url + "/storage/v1/b", json={"name": bucket_name}, verify=False)
        resp.raise_for_status()

        return resp.json()

    def clear_bucket(self, bucket_name: str, prefix: Optional[str] = None) -> Iterable[str]:
        url = self.local_url()
        params = {}
        if prefix:
            params["prefix"] = prefix
        page_token = None
        ret = []
        while True:
            if page_token:
                params["pageToken"] = page_token
            resp = requests.get(url + f"/storage/v1/b/{bucket_name}/o", params=params, verify=False)
            resp.raise_for_status()
            data = resp.json()
            for item in data["items"]:
                ret.append(item["name"])
                requests.delete(url + f'/storage/v1/b/{bucket_name}/o/{item["name"]}', verify=False).raise_for_status()
            if "nextPageToken" in data:
                page_token = data["next_page_token"]
            else:
                break
        return ret

    def delete_bucket(self, bucket_name: str, force: bool = False, missing_ok: bool = False):
        url = self.local_url()

        try:
            if force:
                # we need to delete all the objects in the bucket
                self.clear_bucket(bucket_name)

            resp = requests.delete(url + f"/storage/v1/b/{bucket_name}", verify=False)
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 404 or not missing_ok:
                raise
